/*
 * UNDIGEST
 *
 *     A filter program to split "digest" messages (formatted according
 *     to the de-facto standard, RFC 1153) into their component parts.
 *
 *     (C) Copyright 1991, 1992 Richard B. Wales.  All Rights Reserved.
 *     Permission is granted to use this program for any purpose, so
 *     long as this copyright notice is retained intact, this original
 *     source code (or a modified source with the changes clearly indi-
 *     cated as such) is included as part of any distribution, and no
 *     fee (other than a nominal communications charge) is assessed.
 *
 * DESCRIPTION
 *
 *     This program reads the standard input and writes to the standard
 *     output.  Or, when invoked with a "-d" flag and a user name, it
 *     reads from the standard input and appends its output to the
 *     user's incoming mailbox file in /usr/spool/mail.
 *
 *     The input is expected to be in the UNIX-style mailbox format
 *     (with a "From " line at the start of every message).  Messages
 *     which are not in "digest" format are copied to the standard
 *     output without change.  Digests are split into their component
 *     messages on the standard output.
 *
 *     The "From " line delimiter from the original digest message is
 *     copied unchanged to the start of each component message.  Also,
 *     any "Status:" line in the original header is copied unchanged to
 *     each component message -- replacing any "Status:" line which may
 *     already exist in a component message.
 *
 *     The "To:" line from the original digest message is added to each
 *     component message.  No attempt is made to delete an existing
 *     "To:" line in a component (though ordinarily there will be none).
 *
 *     An "X-Digest:" header line is added to each component message.
 *     This line repeats the "Subject:" information from the digest
 *     header.
 *
 *     When invoked with the "-d" option, the standard UNIX mailbox
 *     locking and "comsat" (asynchronous mail notification) procedures
 *     are adhered to.  Thus, this program may safely be used as a mail
 *     filter in a user's ".forward" file.  For example, the author's
 *     ".forward" file looks like this:
 *
 *                 "| /valeria/wales/bin/undigest -d wales"
 *
 *     This program has been successfully compiled and tested under the
 *     SunOS 4.1.1 operating system and "cc" compiler.  It may require
 *     modification to work in other environments.
 *
 * THINGS TO DO
 *
 *     Extra blank lines seem to get added to the ends of the component
 *     messages within a digest.  It would be nice to fix this.
 *
 *     The user interface could (should?) be changed to make delivery
 *     ("-d username") mode the default.
 */

#include <stdio.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*
 * Digest header info, to be inserted into component messages.
 */
#define LINELEN 1024			/* hopefully adequate */
char	D_unixfrom[LINELEN];		/* "From " line */
char	D_to[LINELEN];			/* "To:" line */
char	D_xdigest[LINELEN];		/* "X-Digest:" line */
char	D_status[LINELEN];		/* "Status:" line */

/*
 * Digest processing status flag.
 */
#define FIRST_LINE	0		/* first line of input */
#define NON_DIGEST_HDR	1		/* non-digest (header) */
#define NON_DIGEST_BODY 2		/* non-digest (body) */
#define DIGEST_PREAMBLE 3		/* digest (preamble) */
#define END_PREAMBLE	4		/* end of preamble seen */
#define COMPONENT_HDR	5		/* component message (header) */
#define COMPONENT_BODY	6		/* component message (body) */
#define END_COMPONENT	7		/* end of component seen */
int	state;


/*
 * Read a line of up to "len-1" characters into a buffer.  The newline
 * character at the end of the line is discarded, and the line is ter-
 * minated by a null character.
 *
 * This routine will end up inserting extra newlines if the incoming
 * text contains null characters or overly long lines.
 */
int
getline (buf, len)
    register char *buf;
    register int len;
{   register int ch;

    if ((ch = getchar ()) == EOF) return -1;
    while (--len > 0 && ch != '\n' && ch != EOF && ch != 0)
    {	*buf++ = ch;
	ch = getchar ();
    }
    if (len <= 0) ungetc (ch, stdin);
    *buf = 0;
    return 0;
}


/*
 * Compare two strings in case-independent fashion.
 * Same as "strncasecmp", which doesn't seem to exist on the Suns.
 */
int
cstrncmp (s1, s2, len)
    register char *s1, *s2;
    register int len;
{   register int n1, n2;

    while (len-- > 0)
    {	n1 = *s1++; n2 = *s2++;
	if (n1 == 0 && n2 == 0) return 0;
	if (n1 >= 'A' && n1 <= 'Z') n1 -= 'A' - 'a';
	if (n2 >= 'A' && n2 <= 'Z') n2 -= 'A' - 'a';
	if (n1 != n2) return n1 - n2;
    }
    return 0;
}

/*
 * Output a null-terminated line, plus a newline.
 */
putline (buf)
    register char *buf;
{   register int ch;

    while ((ch = *buf++) != 0) putchar (ch);
    putchar ('\n');
}

/*
 * Send a message to the "comsat" daemon (asynchronous notification of
 * new mail).  This routine was copied intact from /bin/mail.
 */
notifybiff (msg)
	register char *msg;
{   static struct sockaddr_in addr;
    static int f = -1;

    if (addr.sin_family == 0)
    {	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(IPPORT_BIFFUDP);
    }
    if (f < 0)
	f = socket(AF_INET, SOCK_DGRAM, 0);
    sendto (f, msg, strlen (msg)+1, 0, &addr, sizeof (addr));
}


main (argc, argv)
    int argc;
    char **argv;
{   register int n;
    register char *c;
    static char buffer[LINELEN];
    char biffinfo[20];
    int delivery;

    /* Initialization. */
    state         = NON_DIGEST_HDR;
    D_unixfrom[0] = 0;
    D_to[0]       = 0;
    D_xdigest[0]  = 0;
    D_status[0]   = 0;
    delivery      = 0;
    biffinfo[0]   = 0;

    /*
     * See if there is a "-d" flag (delivery mode).
     * An additional argument after "-d" gives the output file name;
     * or, if it doesn't begin with a slash, the name of the recipient
     * (with an output file in the directory "/usr/spool/mail").
     */
    if (argc >= 2 && strcmp (argv[1], "-d") == 0)
    {	delivery = 1;
	state = FIRST_LINE;
	if (argc >= 3)
	{   if (argv[2][0] != '/')
	    {	strcpy (buffer, "/usr/spool/mail/");
		strcat (buffer, argv[2]);
		strncpy (biffinfo, argv[2], 8);
		biffinfo[8] = 0;
		argv[2] = buffer;
	    }
	    freopen (argv[2], "a", stdout);
	    while (flock (1, LOCK_EX) == EWOULDBLOCK) sleep (5);
	    fseek (stdout, 0L, 2);
    }	}

    /*
     * Save the necessary information so the user can be notified of
     * the new mail's arrival later on via the "comsat" mechanism.
     */
    if (biffinfo[0] != 0)
	sprintf (biffinfo + strlen (biffinfo),
		 "@%ld\n", ftell (stdout));

    /* Process the input. */
    while (getline (buffer, sizeof buffer) == 0)
    {	/*
	 * A UNIX-style "From " line signals a new message.  Save a
	 * copy of the line for possible use before component messages
	 * of a digest.  In "delivery" mode, conceal a "From " line in
	 * the body of the message by prepending a ">" to the line.
	 */
	if (strncmp (buffer, "From ", 5) == 0)
	{   if (!delivery || state == FIRST_LINE)
	    {	strcpy (D_unixfrom, buffer);
		putline (buffer);
		D_to[0] = 0;
		D_xdigest[0] = 0;
		D_status[0] = 0;
		state = NON_DIGEST_HDR;
	    }
	    else
	    {	putchar ('>');
		putline (buffer);
	    }
	    continue;
	}

	/*
	 * In "delivery" mode, if a line begins with one or more ">"s
	 * followed by "From ", prepend another ">".
	 */
	if (delivery && buffer[0] == '>')
	{   for (c = buffer; *c == '>'; c++) ;
	    if (strncmp (c, "From ", 5) == 0)
	    {	putchar ('>');
		putline (buffer);
		continue;
	}   }

	/*
	 * An empty line -- or a line beginning with asterisks or the
	 * words "End of" -- is ignored at the end of a digest preamble
	 * or a component of a digest.  Any other kind of line signals
	 * the start of a new digest component.
	 */
	if (state == END_PREAMBLE || state == END_COMPONENT)
	{   if (buffer[0] == 0
		|| cstrncmp (buffer, "End of ", 7) == 0
		|| strncmp (buffer, "*******", 7) == 0)
		continue;
	    state = COMPONENT_HDR;
	    putchar ('\n');
	    putline (D_unixfrom);
	    /* keep going -- no "continue;" here */
	}

	/*
	 * A "Subject:", "To:", or "Status:" line in the header is
	 * saved for possible future use in splitting up a digest.  In
	 * the case of the "Subject:" line, the keyword is replaced
	 * with the new keyword "X-Digest:".
	 *
	 * An empty line signals the end of the header; and the subject
	 * line is examined to see if this is a digest.
	 */
	if (state == NON_DIGEST_HDR)
	{   if (cstrncmp (buffer, "Subject:", 8) == 0)
	    {	for (c = buffer+8; *c == ' ' || *c == '\t'; c++) ;
		/* note that new keyword is longer than old one */
		buffer[sizeof buffer - 3] = 0;
		strcpy (D_xdigest, "X-Digest: ");
		strcpy (D_xdigest + 10, c);
	    }
	    else if (cstrncmp (buffer, "To:", 3) == 0)
	    {	strcpy (D_to, buffer);
	    }
	    else if (cstrncmp (buffer, "Status:", 7) == 0)
	    {	strcpy (D_status, buffer);
	    }
	    else if (buffer[0] == 0)
	    {	/* is the second word in the line "Digest"? */
		c = D_xdigest + 10;
		while (*c == ' ' || *c == '\t') c++;
		while (*c != 0 && *c != ' ' && *c != '\t') c++;
		while (*c == ' ' || *c == '\t') c++;
		if (cstrncmp (c, "Digest ", 7) == 0)
		     state = DIGEST_PREAMBLE;
		else state = NON_DIGEST_BODY;
	    }
	    putline (buffer);
	    continue;
	}

	/*
	 * The body of a non-digest message is simply passed through.
	 */
	if (state == NON_DIGEST_BODY)
	{   putline (buffer);
	    continue;
	}

	/*
	 * The digest preamble is passed through, until it is ended by
	 * a line consisting of exactly 70 hyphens.
	 */
	if (state == DIGEST_PREAMBLE)
	{   for (n = 70, c = buffer; n > 0 && *c == '-'; n--, c++) ;
	    if (n != 0 || *c != 0) putline (buffer);
	    else                   state = END_PREAMBLE;
	    continue;
	}

	/*
	 * A header line in a digest component is passed through,
	 * except that any "Status:" line is deleted.  After the
	 * end of the component header, a "To:", "X-Digest:", and
	 * "Status:" line are added.
	 */
	if (state == COMPONENT_HDR)
	{   if (buffer[0] == 0)
	    {	if (D_to[0] != 0) putline (D_to);
		putline (D_xdigest);
		if (D_status[0] != 0) putline (D_status);
		putchar ('\n');
		state = COMPONENT_BODY;
	    }
	    else if (cstrncmp (buffer, "Status:", 7) != 0)
		putline (buffer);
	    continue;
	}

	/*
	 * The component body is passed through, until it is ended by
	 * a line consisting of exactly 30 hyphens.
	 */
	if (state == COMPONENT_BODY)
	{   for (n = 30, c = buffer; n > 0 && *c == '-'; n--, c++) ;
	    if (n != 0 || *c != 0) putline (buffer);
	    else                   state = END_COMPONENT;
	    continue;
	}

	/*NOTREACHED*/
    }

    /* Windup. */
    if (delivery) putchar ('\n');	/* just to be sure */

    /* Use the "comsat" facility to notify the mail's recipient. */
    notifybiff (biffinfo);

    /* All done. */
    exit (0);
}
